#
# Copyright (C) 2019-2020 Guillaume Bernard <guillaume.bernard@koala-lms.org>
# Copyright (C) 2020 Raphaël Penault <raphael.penault@etudiant.univ-lr.fr>
#
# This file is part of Koala LMS (Learning Management system)

# Koala LMS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# We make an extensive use of the Django framework, https://www.djangoproject.com/

import json

from django.contrib.auth.hashers import check_password
from django.core import mail
from django.test import TestCase, Client
from django.urls import reverse, reverse_lazy
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode

import accounts
from accounts.forms import UserCreationForm
from accounts.models import Person, Notification
from accounts.tokens import AccountsActivationTokenGenerator
from accounts.views import send_activation_mail


class MarkNotificationAsReadTestView(TestCase):

    def setUp(self) -> None:
        self.user = Person.objects.create_user(id=1, username="william-shakespeare", password="pwd")
        self.client = Client()
        self.client.login(username="william-shakespeare", password="pwd")
        Notification.objects.create(id=1, recipient=self.user, message="A test", target="https://google.com")
        Notification.objects.create(id=2, recipient=self.user, message="A test", target="https://google.com")
        Notification.objects.create(id=3, recipient=self.user, message="A test 2", target="https://google.com")
        Notification.objects.create(id=4, recipient=self.user, message="A test 3")
        Notification.objects.create(id=5, recipient=self.user, message="A test 4", target="https://google.com")
        Notification.objects.create(id=6, recipient=self.user, message="A test 5")
        Notification.objects.create(id=7, recipient=self.user, message="A test 6", target="https://google.com")

    def test_get_mark_notification_as_read(self):
        self.assertEqual(7, self.user.unread_notifications.count())
        response = self.client.get(
            reverse("accounts:ajax/notification/read", kwargs={'notification_id': 1})
        )
        self.assertEqual(405, response.status_code)
        self.assertEqual(7, self.user.unread_notifications.count())

    def test_post_mark_notification_as_read(self):
        self.assertEqual(7, self.user.unread_notifications.count())
        response = self.client.post(
            reverse("accounts:ajax/notification/read", kwargs={'notification_id': 1})
        )
        self.assertEqual(200, response.status_code)
        json_object = json.loads(response.content)
        self.assertEqual(6, self.user.unread_notifications.count())
        self.assertEqual(6, json_object.get('unread'))

    def test_get_delete_notification(self):
        self.assertEqual(7, self.user.unread_notifications.count())
        response = self.client.get(
            reverse("accounts:ajax/notification/delete", kwargs={'notification_id': 1})
        )
        self.assertEqual(405, response.status_code)
        self.assertEqual(7, self.user.unread_notifications.count())

    def test_post_delete_notification(self):
        self.assertEqual(7, self.user.unread_notifications.count())
        response = self.client.post(
            reverse("accounts:ajax/notification/delete", kwargs={'notification_id': 1})
        )
        self.assertEqual(200, response.status_code)
        json_object = json.loads(response.content)
        self.assertEqual(6, self.user.unread_notifications.count())
        self.assertEqual(6, json_object.get('unread'))


class TestUserLoginView(TestCase):

    def setUp(self) -> None:
        self.user = Person.objects.create_user(id=1, username="william-shakespeare", password="good-password",
                                               is_active=False)
        self.client = Client()

    def test_user_login_form_with_anonymous_username(self):
        """
        Test the login process with anonymous as login, and a random password. This test exists because some raised a
        runtime error when choosing anonymous as login.
        """
        response = self.client.post(
            reverse("accounts:login"), {"username": "anonymous", "password": ""}
        )
        user = response.context["user"]
        self.assertEqual(200, response.status_code)
        self.assertTemplateUsed(response, "registration/login.html")
        self.assertContains(response, "Login unsuccessful")
        self.assertTrue(user.is_anonymous)
        self.assertFalse(response.context["user"].is_authenticated)

    def test_user_login_form_with_anonymous_username_and_password(self):
        """
        Test the login process with anonymous as login, and a random password. This test exists because some raised a
        runtime error when choosing anonymous as login.
        """
        response = self.client.post(
            reverse("accounts:login"), {"username": "anonymous", "password": "password"}
        )
        user = response.context["user"]
        self.assertEqual(200, response.status_code)
        self.assertTemplateUsed(response, "registration/login.html")
        self.assertContains(response, "Login unsuccessful")
        self.assertTrue(user.is_anonymous)
        self.assertFalse(response.context["user"].is_authenticated)

    def test_user_login_form_with_good_password(self):
        """
        Test the login process with a an existing login, and the good password
        """
        self.user.is_active = True
        self.user.save()
        response = self.client.post(
            reverse("accounts:login"), {"username": self.user.username, "password": "good-password"}
        )
        self.assertRedirects(response, reverse("accounts:details"))

    def test_user_login_form_with_wrong_password(self):
        """
        Test the login process with a an existing login, and a wrong password
        """
        response = self.client.post(
            reverse("accounts:login"), {"username": self.user.username, "password": "wrong-password"}
        )
        user = response.context["user"]
        self.assertEqual(200, response.status_code)
        self.assertTemplateUsed(response, "registration/login.html")
        self.assertContains(response, "Login unsuccessful")
        self.assertTrue(user.is_anonymous)
        self.assertFalse(response.context["user"].is_authenticated)

    def test_user_login_valid_user_not_active_resend_message(self):
        self.assertFalse(self.user.is_active)
        response = self.client.post(
            reverse("accounts:login"), {"username": self.user.username, "password": "good-password"}
        )
        self.assertEqual(200, response.status_code)
        self.assertTemplateUsed(response, "registration/login.html")

        # No user is connected, even with valid credentials
        user = response.context["user"]
        self.assertTrue(user.is_anonymous)
        self.assertFalse(response.context["user"].is_authenticated)

        # Check presence of resend confirmation link
        uidb64_username = urlsafe_base64_encode(force_bytes(self.user.username))
        self.assertContains(response, "Another one will be resend to you.")
        self.assertContains(
            response, reverse_lazy('accounts:resend_email_confirmation', kwargs={'uidb64_username': uidb64_username})
        )


class TestResendActivationMail(TestCase):

    def setUp(self) -> None:
        self.user = Person.objects.create_user(
            id=1, username="william-shakespeare", password="good-password",
            email="ws@domain.tld", is_active=False
        )
        self.client = Client()

    def test_resend_activation_email(self):
        uidb64_username = urlsafe_base64_encode(force_bytes(self.user.username))
        response = self.client.get(
            reverse_lazy("accounts:resend_email_confirmation", kwargs={'uidb64_username': uidb64_username})
        )
        self.assertRedirects(response, reverse_lazy("accounts:login"))
        response = self.client.get(
            reverse_lazy("accounts:login")
        )
        self.assertEqual(200, response.status_code)
        self.assertTemplateUsed("registration/login.html")

        token = AccountsActivationTokenGenerator().make_token(self.user)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].subject, "Enable your Koala LMS account")
        self.assertIn(token, mail.outbox[0].body)

    def test_resend_activation_email_without_email_address(self):
        user_no_email = Person.objects.create_user(
            id=1, username="leon-tolstoi", password="good-password", is_active=False
        )

        with self.assertRaises(accounts.exc.PersonWithoutEmailError):
            send_activation_mail(user_no_email, domain="host")

        uidb64_username = urlsafe_base64_encode(force_bytes(user_no_email.username))
        response = self.client.get(
            reverse_lazy("accounts:resend_email_confirmation", kwargs={'uidb64_username': uidb64_username})
        )
        self.assertRedirects(response, reverse("accounts:login"))
        self.assertEqual(len(mail.outbox), 0)


class TestUserRegistrationView(TestCase):

    def setUp(self) -> None:
        self.client = Client()

    def test_user_registration_get_form(self):
        response = self.client.get(reverse_lazy("accounts:register"))
        self.assertEqual(200, response.status_code)
        self.assertTemplateUsed(response, "accounts/register.html")
        for field in UserCreationForm.Meta.fields:
            self.assertContains(response, field)

    def test_user_registration_post_default_values(self):
        user_dict = {
            "first_name": "Firstname",
            "last_name": "Lastname",
            "preferred_language": "en",
            "username": "Username",
            "email": "firstname.lastname@domain.tld",
            "password1": "A@Wr/on|g\Password",
            "password2": "A@Wr/on|g\Password",
        }
        self.assertEqual(Person.objects.count(), 0)
        response = self.client.post(
            reverse("accounts:register"), user_dict
        )
        self.assertRedirects(response, reverse_lazy("accounts:register"))

        # Check context user info
        user = response.context["user"]
        self.assertEqual(Person.objects.count(), 1)
        self.assertEqual(user, Person.objects.get(pk=1))
        self.assertTrue(check_password("A@Wr/on|g\Password", user.password))
        self.assertFalse(user.is_anonymous)
        self.assertTrue(response.context["user"].is_authenticated)
        self.assertFalse(user.is_active)
        for key in ['first_name', 'last_name', 'username', 'email']:
            self.assertEqual(getattr(user, key), user_dict.get(key))

        # Check confirmation email is sent
        uid = urlsafe_base64_encode(force_bytes(user.pk))
        token = AccountsActivationTokenGenerator().make_token(user)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].subject, "Enable your Koala LMS account")
        self.assertIn(reverse("accounts:activate", kwargs={'uidb64': uid, 'token': token}), mail.outbox[0].body)

    def test_activate_accounts_valid_link(self):
        user = Person.objects.create_user(
            id=1, username="william-shakespeare", password="good-password",
            email="ws@domain.tld", is_active=False
        )
        self.assertFalse(user.is_active)
        uid = urlsafe_base64_encode(force_bytes(user.pk))
        token = AccountsActivationTokenGenerator().make_token(user)
        response = self.client.get(
            reverse("accounts:activate", kwargs={'uidb64': uid, 'token': token})
        )
        self.assertRedirects(response, reverse_lazy("accounts:login"))
        self.assertTrue(Person.objects.get(id=1).is_active)

    def test_activate_accounts_invalid_link(self):
        user = Person.objects.create_user(
            id=1, username="william-shakespeare", password="good-password",
            email="ws@domain.tld", is_active=False
        )
        self.assertFalse(user.is_active)
        uid = urlsafe_base64_encode(force_bytes(user.pk)) + '-invalid'
        token = AccountsActivationTokenGenerator().make_token(user)
        response = self.client.get(
            reverse("accounts:activate", kwargs={'uidb64': uid, 'token': token})
        )
        self.assertRedirects(response, reverse_lazy("accounts:login"))
        self.assertFalse(Person.objects.get(id=1).is_active)


class TestViews(TestCase):

    def setUp(self) -> None:
        self.ws = Person.objects.create_user(
            username="william-shakespeare", password="password",
            first_name="William", last_name="Shakespeare"
        )
        self.lt = Person.objects.create_user(id=2, username="leon-tolstoi")
        self.gm = Person.objects.create_user(id=3, username="guy-maupassant")
        self.bw = Person.objects.create_user(
            username="boby-willy", password="password",
            first_name="Boby", last_name="Willy"
        )

        self.client_ws = Client()
        self.client_ws.login(username="william-shakespeare", password="password")

    def test_search_user_unauthenticated_user(self):
        self.client = Client()
        response = self.client.get(
            reverse("accounts:ajax/search"), {'user': 'wil'}
        )
        self.assertRedirects(
            response, reverse("accounts:login") + "?next=" + reverse("accounts:ajax/search") + "%3Fuser%3Dwil"
        )

    def test_search_user_authenticated_user(self):
        response = self.client_ws.get(
            reverse("accounts:ajax/search"), {'user': 'willi'}
        )
        self.assertEqual(200, response.status_code)
        user_list = json.loads(response.content)
        self.assertEqual(1, len(user_list))
        ws_dict = user_list[0]
        self.assertEqual("william-shakespeare", ws_dict.get('username'))
        self.assertEqual("William", ws_dict.get('first_name'))
        self.assertEqual("Shakespeare", ws_dict.get('last_name'))

    def test_search_users_authenticated_user(self):
        response = self.client_ws.get(
            reverse("accounts:ajax/search"), {'user': 'Wil'}
        )
        self.assertEqual(200, response.status_code)
        user_list = json.loads(response.content)

        # Users are ordered according to their usernames
        bw_dict = user_list[0]
        self.assertEqual(self.bw.username, bw_dict.get('username'))
        self.assertEqual(self.bw.first_name, bw_dict.get('first_name'))
        self.assertEqual(self.bw.last_name, bw_dict.get('last_name'))
        ws_dict = user_list[1]
        self.assertEqual(self.ws.username, ws_dict.get('username'))
        self.assertEqual(self.ws.first_name, ws_dict.get('first_name'))
        self.assertEqual(self.ws.last_name, ws_dict.get('last_name'))
