#
# Copyright (C) 2019-2020 Guillaume Bernard <guillaume.bernard@koala-lms.org>
# Copyright (C) 2020 Louis Barbier <louis.barbier41@outlook.fr>
#
# This file is part of Koala LMS (Learning Management system)

# Koala LMS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# We make an extensive use of the Django framework, https://www.djangoproject.com/
from django.contrib import messages
from django.contrib.auth import get_user_model
from django.contrib.auth.decorators import login_required
from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib.auth.views import LoginView as AuthLoginView, PasswordChangeView
from django.core.mail import EmailMessage
from django.db.models import Q
from django.http import JsonResponse, HttpResponseRedirect, HttpRequest
from django.shortcuts import get_object_or_404, redirect
from django.template.loader import render_to_string
from django.urls import reverse_lazy, reverse
from django.utils.encoding import force_bytes, force_text
from django.utils.http import urlsafe_base64_encode, urlsafe_base64_decode
from django.utils.translation import ugettext_lazy as _
from django.views.decorators.http import require_http_methods
from django.views.generic import CreateView, UpdateView

import accounts.exc
from accounts.forms import UserLoginForm, UserCreationForm, UserChangeForm
from accounts.models import Person, Notification
from accounts.tokens import AccountsActivationTokenGenerator


def _update_valid_or_invalid_form_fields(form):
    for field in form.fields:
        try:
            current_class = form.fields[field].widget.attrs['class']
        except KeyError:
            current_class = str()

        if field in form.errors:
            form.fields[field].widget.attrs.update({'class': current_class + ' ' + 'is-invalid'})
        elif field in form.changed_data:
            form.fields[field].widget.attrs.update({'class': current_class + ' ' + 'is-valid'})
    return form


def send_activation_mail(user: get_user_model(), domain: str) -> bool:
    """
    This function sends an activation email. It can be used to send the first confirmation email or to send a new one

    :param user: This contains the user who wants activate his account
    :type user: get_user_model()
    :param domain: the domain on which the current instance is running. This is, for instance: demo.koala-lms.org
    :type domain: str
    :return True if the mail has been send
    :rtype bool
    """
    if not user.email:
        raise accounts.exc.PersonWithoutEmailError(
            _("Person %(person)s does not have any email address registered in the profile. "
              "Sending email is impossible.") % {'person': user}
        )
    reverse_url = reverse(
        'accounts:activate',
        kwargs={
            'uidb64': urlsafe_base64_encode(force_bytes(user.pk)),
            'token':AccountsActivationTokenGenerator().make_token(user)
        }
    )
    message = render_to_string('registration/account_activation_mail.html', {
        'user': user,
        'domain': domain,
        'url': f"http://{domain}{reverse_url}"
    })
    message_sent = EmailMessage(_("Enable your Koala LMS account"), message, to=[user.email]).send() == 1
    if not message_sent:
        raise accounts.exc.CannotSendEmailError(
            _("The component used to send the confirmation mails is broken. Please contact your "
              "system administrator to let him or her fix this issue.")
        )
    return message_sent


@require_http_methods(['GET'])
def resend_activation_mail(request, uidb64_username: str):
    """
    This view is used to resent the confirmation email to the user that requested it.

    :param request: the request to handle
    :type request: django.http.HttpRequest
    :param uidb64_username: the user’s username to send the confirmation email. username is encoded using urlsafe_base64_encode
    :type uidb64_username: str
    """
    try:
        decoded_user = get_object_or_404(get_user_model(), username=force_text(urlsafe_base64_decode(uidb64_username)))
    except(TypeError, ValueError, OverflowError, get_user_model().DoesNotExist):
        messages.error(
            request,
            _("Something went wrong when trying to find your account in the database. Aborting.")
        )
    else:
        try:
            if send_activation_mail(decoded_user, request.get_host()):
                messages.info(
                    request, _(
                        "%(user)s, we send you a new confirmation email to %(email)s."
                    ) % {'user': decoded_user.get_full_name(), 'email': decoded_user.email}
                )
        except (accounts.exc.PersonWithoutEmailError, accounts.exc.CannotSendEmailError) as ex:
            messages.error(request, str(ex))
    return redirect("accounts:login")


class LoginView(AuthLoginView):
    """
    The view that handles login actions.
    """
    form_class = UserLoginForm
    redirect_authenticated_user = True

    def form_invalid(self, form):
        """
        This function is called when the UserLoginForm is valid. In this method, we process the form data in order to
        know if the user that wanted to authenticate is active or not. If it is not, a reason might be that its account
        has not been explicitly activated.

        .. note:: FIXME: In case the administrator disabled the user explicitly, the user will receive an email that
                  will allow him/her to re-activate the account.

        :param form: the form that has been processed
        :type form: UserLoginForm
        :return: the return value of parent form_invalid(…) call.
        """
        # form.user_cache is here the user that is, actually authenticated.
        if form.user_cache and not form.user_cache.is_active:
            uidb64_username = urlsafe_base64_encode(force_bytes(form.user_cache.username))
            messages.warning(
                self.request,
                _("%(name)s, is seems your account has not been activated yet. If you did not received any "
                  "confirmation email, click <a href=\"%(resend_link)s\">here</a>. Another one will be resend "
                  "to you.")
                % {
                    'name': form.user_cache.get_full_name(),
                    'resend_link': reverse_lazy(
                        'accounts:resend_email_confirmation',
                        kwargs={'uidb64_username': uidb64_username}
                    )
                }
            )
        else:
            messages.error(self.request, _("Login unsuccessful. Please verify your username and password."))
        return super().form_invalid(form)


class PasswordChange(PasswordChangeView):
    """
    The view that is responsible of updating the user password, upon request.
    """
    success_url = reverse_lazy('accounts:details')
    title = _("My account − Change password")

    # noinspection PyMissingOrEmptyDocstring
    def form_valid(self, form):
        messages.success(self.request, _("Your password has been updated!"))
        return super().form_valid(form)

    # noinspection PyMissingOrEmptyDocstring
    def form_invalid(self, form):
        messages.error(
            self.request, _("Error when updating password. You should fix the issues in the form and try again.")
        )
        return super().form_invalid(_update_valid_or_invalid_form_fields(form))


class AccountsRegisterView(CreateView):
    """
    The view using which users can register on the application
    """
    template_name = "accounts/register.html"
    form_class = UserCreationForm
    success_url = reverse_lazy("accounts:register")
    extra_context = {
        "title": _("Create a new account")
    }

    def form_valid(self, form):
        """
        In case the form is valid, save it, disable it, and save it in database. An activation email is sent to him.

        :param form: the form used to create a new user
        :type form: UserCreationForm
        :return: the return value of parent form_valid(…) call.
        """
        user = form.save()
        user.is_active = False  # disable user by default, user needs to explicitly use an activation link
        user.save()
        try:
            if send_activation_mail(user, self.request.get_host()):
                messages.success(
                    self.request, _(
                        "%(user)s (%(username)s), your registration is complete now. In order to access your account, you "
                        "have to activate it first. We just sent you an email to confirm your registration."
                    ) % {'user': user.get_full_name(), 'username': user.username}
                )
        except (accounts.exc.PersonWithoutEmailError, accounts.exc.CannotSendEmailError) as ex:
            messages.error(self.request, ex.message)
        return super().form_valid(form)

    def form_invalid(self, form):
        """
        In case the form is invalid, errors are shown to the user.

        :param form: the form used to create a new user
        :type form: UserCreationForm
        :return: the return value of parent form_invalid(…) call.
        """
        for error in form.errors:
            messages.error(self.request, form.errors[error])
        return super().form_invalid(form)


class AccountsDetailsView(LoginRequiredMixin, UpdateView):
    """
    View to show account details.
    """
    template_name = "accounts/details.html"
    form_class = UserChangeForm
    success_url = reverse_lazy("accounts:details")

    def get_object(self, queryset=None):
        """
        The object that is processed here is the logged-in user.
        :return: the current, logged in user.
        :rtype: Person
        """
        return self.request.user

    # noinspection PyMissingOrEmptyDocstring
    def form_valid(self, form):
        form.save()
        if form.has_changed():
            messages.success(self.request, _("Your personal details have been updated!"))
        return super().form_valid(form)

    # noinspection PyMissingOrEmptyDocstring
    def form_invalid(self, form):
        messages.error(self.request, _("Error when updating personal details. Check for errors and retry."))
        return super().form_invalid(_update_valid_or_invalid_form_fields(form))

    # noinspection PyMissingOrEmptyDocstring
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["title"] = _("My account − %(username)s") % {'username': self.request.user.display_name}
        return context


@login_required
@require_http_methods(['GET'])
def search_user(request: HttpRequest) -> JsonResponse:
    """
    Searching for a user is limited to − obviously − registered user.
    It seems that this may be a personal data leak, but as any registered user can create a course and add collaborator
    or students, up to now there is not other solution.

    :param request: the current request object
    :type request: django.http.HttpRequest
    :return: the view response
    :rtype: django.http.JsonResponse
    """
    response = JsonResponse({})
    search_string = request.GET.get("user", None)
    if search_string is not None and len(search_string) > 1:
        people = Person.objects.filter(
            Q(username__icontains=search_string) |
            Q(first_name__icontains=search_string) |
            Q(last_name__icontains=search_string)
        )
        response = JsonResponse(
            list(people.values('username', 'first_name', 'last_name')),
            safe=False
        )
    return response


@login_required
@require_http_methods(['POST'])
def notification_mark_as_read(request: HttpRequest, notification_id: int) -> JsonResponse:
    """
    Update the “is_read” attribute of a Notification object to set the notification as read.

    :param request: the current request object
    :param notification_id: the notification is the current user owns to set as read
    :return: a json object with the current unread notification count
    :rtype: django.http.JsonResponse
    """
    notification = get_object_or_404(Notification, recipient=request.user, pk=notification_id)
    notification.is_read = True
    notification.save()
    return JsonResponse({'unread': request.user.unread_notifications.count()})


@login_required
@require_http_methods(['POST'])
def notification_delete(request: HttpRequest, notification_id: int) -> JsonResponse:
    """
    Delete a notification object.

    :param request: the current request object
    :param notification_id: the notification id the current user owns to delete
    :return: a json object with the current unread notification count
    :rtype: django.http.JsonResponse
    """
    notification = get_object_or_404(Notification, recipient=request.user, pk=notification_id)
    notification.delete()
    return JsonResponse({'unread': request.user.unread_notifications.count()})


@require_http_methods(['GET'])
def activate(request: HttpRequest, uidb64: str, token: str) -> HttpResponseRedirect:
    """
    Activate the current user’s account using a specific token

    :param request: the current request object
    :type request: django.http.HttpRequest
    :param uidb64: the base 64 encoded string containing the user id
    :type uidb64: str
    :param token: the activation token sent to the user by email
    :type token: str
    :return: a redirect response to the login page
    :rtype: HttpResponseRedirect
    """
    try:
        user = get_object_or_404(Person, pk=force_text(urlsafe_base64_decode(uidb64)))
    except(TypeError, ValueError, OverflowError, Person.DoesNotExist):
        messages.error(
            request,
            _("The activation link you provided is broken, or does not correspond to something expected. Please check "
              "your email in order to find your activation link.")
        )
    else:
        if AccountsActivationTokenGenerator().check_token(user, token):
            user.is_active = True
            user.save()
            messages.success(
                request,
                _("%(user)s (%(username)s), your account has been confirmed. You can now log in using the password "
                  "you previously chose.") % {'user': user.get_full_name(), 'username': user.username}
            )
    return redirect("accounts:login")
